/**
***	$RCSfile: hd.c,v $
***	$Revision: 1.13 $
***	$Date: 2010/01/05 16:23:55 $ UTC (aka GMT)
***	$Author: adamm $
**/


#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#define		CMDLINE		"[ options ] [ file ... ]"
#define		DOT		'.'
#define		SPC		' '


static char RCSid[] =
    "$Id: hd.c,v 1.13 2010/01/05 16:23:55 adamm Exp adamm $";


void		fillHiLo	(void);
void		fillTransTbl	(int);
void		manpage		(void);
void		processArgs	(int, char *[]);
int		showFile	(char *);
void		showLine	(off_t addr, unsigned char *, int);
void		unhexify	(unsigned char **);
void		usage		(void);


char		AddrFmt[16]		= "%0*jx";
int		AddrWidth		= 8;
char		*AscBuf;
int		AscBufLen;
int		BlockGutter		= 3;
int		BytesPerBlock		= 0;
int		BytesPerGroup		= 1;
int		BytesPerLine		= 16;
int		Delim			= 0;
int		GroupGutter		= 1;
char		*HexBuf;
int		HexBufLen;
char		HexChars[]		= "0123456789abcdef";
unsigned char	Hi[256];
unsigned char	LeftDelimChar		= '\0';
int		LeftGutter		= 2;
unsigned char	Lo[256];
int		NonPrintChar		= (int) '.';
int		PrintBlanks		= 0;
char		*Prog;
unsigned char	RightDelimChar		= '\0';
int		RightGutter		= 2;
unsigned char	TransTbl[256];
int		Verbose			= 0;


int main(int argc, char *argv[])
{
    int r = 0;

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	Prog = &argv[0][0];
    } else {
	++Prog;
    }

    /* Parse the command line */
    processArgs(argc, argv);

    if ((Delim == 1) && (LeftDelimChar == '\0')) {
	LeftDelimChar = RightDelimChar = '|';
    }

    /* Fill the hex table */
    fillHiLo();

    /* Fill the translation table */
    fillTransTbl(NonPrintChar);
    if (PrintBlanks) {
	TransTbl[(int) ' '] = ' ';
    }

    /* Calculate the size of the hex buffer */
    AscBufLen = BytesPerLine + 3;
    HexBufLen = 10 + AddrWidth + LeftGutter + RightGutter +
	(BytesPerLine * 2) +
	((ceil(1.0 * BytesPerLine / BytesPerGroup) - 1) * GroupGutter) +
	((BytesPerBlock > 0) ?
	    ((ceil(1.0 * BytesPerLine / BytesPerBlock) - 1) * BlockGutter)
	    : 0) +
	AscBufLen;

    /**
    *** Why 10?
    ***
    *** To make room for the colon, the NL and NUL at the end of the string,
    *** and 8 bytes of slop (mostly to allow for address overflow).
    ***
    *** This will always be larger than we need, since BlockGutter
    *** supercedes some instances of GroupGutter, but this calculation is
    *** significantly less costly.
    **/

    /* Create the hex and ASCII buffers */
    AscBuf = (char *) malloc(AscBufLen);
    if (AscBuf == NULL) {
	(void) fprintf(stderr, "%s: malloc() failed!\n", Prog);
	exit(1);
    }

    HexBuf = (char *) malloc(HexBufLen);
    if (HexBuf == NULL) {
	(void) fprintf(stderr, "%s: malloc() failed!\n", Prog);
	exit(1);
    }

    /* Read the file(s) and/or stdin */
    if (optind == argc) {
	r = r | showFile("");

	/**
	*** We use r = r | ... because POSIX.2 sez we're supposed to
	*** print a warning message, keep going, then exit with a
	*** non-zero value. So there.
	**/

    } else {
	for ( ; optind < argc; optind++) {
	    if (strcmp(argv[optind], "-") == 0) {
		r = r | showFile("");
	    } else {
		r = r | showFile(argv[optind]);
	    }
	}
    }

    exit(r);
} /* main */


void fillHiLo (void)
{
    int i;

    for (i = 0; i < 256; i++) {
	Hi[i] = HexChars[i >> 4];
	Lo[i] = HexChars[i & 0x0f];
    }

    return;
} /* fillHiLo */


void fillTransTbl(int c)
{
    int i;

    (void) memset(TransTbl, c, 256);

    for (i = 33; i < 127; i++) {
	TransTbl[i] = (unsigned char) i;
    }

    return;
} /* fillTransTbl */


void manpage(void)
{
	unsigned char *lines[] = {
	    #include "hd.man.hex"
	    0
	};

	unhexify(lines);
} /* manpage */


void processArgs(int ac, char *av[])
{
    int  c;
    int  colon = 0;
    size_t len;
    char *p;

    /* Special handling for --help and --man */
    if (ac > 1) {
	if (
	    (strcmp(av[1], "--help") == 0)
	    ||
	    (strcmp(av[1], "--man") == 0)
	) {
	    manpage();
	    exit(0);
	}
    }

    while ((c = getopt(ac, av, ":A:B:bC:D:dG:L:N:R:uv")) != -1) {
	switch (c) {
	case 'A':
	    if (isdigit(*optarg) != 0) {
		AddrWidth = atoi(optarg);
	    }

	    p = strchr(optarg, ':');
	    if (p != NULL) {
		++colon;
	    }
	    break;

	case 'B':
	    BytesPerBlock = atoi(optarg);
	    p = strchr(optarg, DOT);
	    if (p != NULL) {
		BlockGutter = atoi(++p);
	    }
	    break;

	case 'b':
	    PrintBlanks = 1;
	    Delim = 1;
	    break;

	case 'C':
	    NonPrintChar = (*optarg);
	    break;

	case 'D':
	    len = strlen(optarg);
	    switch (len) {
	    case 1:
		LeftDelimChar = RightDelimChar = *optarg;
		Delim = 1;
		break;
	    case 2:
		LeftDelimChar = *optarg;
		RightDelimChar = *(optarg + 1);
		Delim = 1;
		break;
	    default:
		(void) fprintf(stderr, "%s: bad delimiter(s): '%s'\n",
		    Prog, optarg);
		exit(1);
	    }
	    break;

	case 'd':
	    Delim = 1;
	    break;

	case 'G':
	    BytesPerGroup = atoi(optarg);
	    p = strchr(optarg, DOT);
	    if (p != NULL) {
		GroupGutter = atoi(++p);
	    }
	    if (BytesPerGroup <= 0) {
		BytesPerGroup = 1;
	    }
	    break;

	case 'L':
	    LeftGutter = atoi(optarg);
	    break;

	case 'N':
	    BytesPerLine = atoi(optarg);
	    break;

	case 'R':
	    RightGutter = atoi(optarg);
	    break;

	case 'u':
	    (void) strcpy(AddrFmt, "%0*jX");
	    (void) strcpy(HexChars + 10, "ABCDEF");
	    break;

	case 'v':
	    Verbose = 1;
	    break;

	case '?':
/** FIX THIS **/
	    (void) fprintf(stderr,"bad arg: %c\n",optopt);
	    usage();
	    exit(1);

	default:
	    usage();
	    exit(1);
	}
    }

    /* If necessary, mung AddrFmt */
    if (colon > 0) {
	(void) strcat(AddrFmt, ":");
    }


    return;
} /* processArgs */


int showFile(char *fileName)
{
    off_t		addr;
    int			bytes;
    unsigned char	buf[BUFSIZ];
    FILE		*fp;
    unsigned char	last[32];
    int			n;
    int			p;
    char		same[4];

    /* Open the file (or die trying) */
    if (strcmp(fileName, "") != 0) {
	fp = fopen(fileName, "rb");
	if (fp == NULL) {
	    (void) fprintf(stderr,
		    "%s:%d: Unable to open file \"%s\": %s\n",
		    Prog, errno, fileName, strerror(errno));
	    return(1);
	}
    } else {
	fp = stdin;
    }

    /* address is relative to *this* file */
    addr = (off_t) 0;

    /* Fill last[] with an unlikely value */
    (void) memset(last, 0xE5, 32);

    /* Clear same[] */
    (void) memset(same, 0, 4);

    /* Process the file, showing each "line" */
    p = 0;
    do {
	/* Read a big chunk of data */
	n = fread(buf + p, sizeof(char), (size_t) (BUFSIZ - p), fp);

	if (ferror(fp) > 0) {
	    (void) fprintf(stderr, "%s:%d: read error: %s\n",
		    Prog, errno, strerror(errno));
	    return(1);

	} else if (n > 0) {
	    /* Process it one line at a time */
	    do {
		/* Full line or partial? */
		if ((n - p) < BytesPerLine) {
		    bytes = n - p;
		} else {
		    bytes = BytesPerLine;
		}

		/* Show 'em all */
		if (Verbose == 1) {
		    showLine(addr, buf + p, bytes);

		} else {
		    /* Test for duplicate lines */
		    if (memcmp(buf + p, last, bytes) == 0) {
			(void) strcpy(same, "*\n");
		    } else {
			(void) printf("%s", same);
			showLine(addr, buf + p, bytes);
			(void) memset(same, 0, 4);
			(void) memcpy(last, buf + p, bytes);
		    }
		}

		addr += (off_t) bytes;
		p += bytes;
	    } while (p < n);

	    /* Relocate unprocessed bytes */
	    if (p < n) {
		(void) memmove(buf, buf + p, n - p);
		p = n - p;
	    } else {
		p = 0;
	    }
	}
    } while (feof(fp) == 0);

    /* Print any left-overs */
    if (p > 0) {
	showLine(addr, buf, p);
    }

    (void) fclose(fp);
    return(0);
} /* showFile */


void showLine(off_t addr, unsigned char *bp, int n)
{
    char *ap = AscBuf;
    char *hp = HexBuf;
    int  i;

    /* Clear the buffers */
    (void) memset(ap, SPC, AscBufLen);
    (void) memset(hp, SPC, HexBufLen);

    if (AddrWidth > 0) {
	hp += sprintf(hp, AddrFmt, AddrWidth, (uintmax_t) addr);
	*hp = SPC;

	hp += LeftGutter;
    }

    if (Delim > 0) {
	*ap++ = LeftDelimChar;
    }

    for (i = 0; i < BytesPerLine; i++) {
	if ((BytesPerBlock > 0) &&
		((i % BytesPerBlock) == 0) && (i > 0)) {
	    hp += BlockGutter;
	
	} else if (((i % BytesPerGroup) == 0) && (i > 0)) {
	    hp += GroupGutter;
	}

	if (i >= n) {
	    hp += 2;
	    ap++;
	} else {
	    *hp++ = Hi[*bp];
	    *hp++ = Lo[*bp];
	    *ap++ = TransTbl[*bp];
	    bp++;
	}
    }

    hp += RightGutter;

    if (Delim > 0) {
	*ap++ = RightDelimChar;
    }

    *ap = '\0';
    *hp = '\0';

    strcat(HexBuf, AscBuf);
    (void) puts(HexBuf);

    return;
} /* showLine */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            putc((int) c, stdout);
        }
    }
} /* unhexify */


void usage(void)
{
    (void) fprintf(stderr, "Usage: %s %s\n", Prog, CMDLINE);
    (void) fprintf(stderr, "\n");
    (void) fprintf(stderr, "Try '%s --man' for more details.\n", Prog);
    return;
} /* usage */



/*	===== MAN PAGE STARTS HERE =====
NAME
    hd -- display a file in hexadecimal and ASCII

SYNOPSIS
    hd [ options ] [ file ... ]

DESCRIPTION
    hd is similar to od(1) (dump files to stdout in various formats),
    but emulates the (very) old IBM "dump" command for output, namely:

	address   bytes-in-hex   bytes-in-ASCII

    Like the POSIX version of od(1), hd has more options than one might
    expect; fortunately, in most cases, the default output is all that
    is needed.

    If no files are given on the command line, or if a dash ("-") is
    given as a filename, hd will read from stdin. If mutiple file names
    are given, there will be no delimiter between the files (in the
    output) but the addresses will start at zero for each file.

BASIC OPTIONS
    -b
	treat blanks as printing characters (implies -d)

    -d
	print delimiters around the ASCII column

    -u
	use upper-case hex characters

    -v
	do not supress consecutive, identical lines

    --help
    --man
	this man page

EXTENDED OPTIONS
    -A a[:]
	set the address width to "a" digits (default is 8); to suppress
	the address column, use "-A 0"; to add a colon (":") after the
	address, use "-A :" or "-A a:"

    -B B[.b]
	show blocks of "B" bytes together, with extra space between
	blocks (no default); if ".b" is given, put "b" spaces between
	groups (default is 3 spaces)

    -C c
	show non-printing characters as "c" (default is ".")

    -D c | cc
	set the delimiter on either side of the column of ASCII bytes to
	be the character specified by "c"; if two characters are given,
	the first character will be used on the left of thecolumn, the
	second character on the right; implied the "-d" flag (above)

    -G G[.g]
	show "G" bytes contiguously (default is 1); if ".g" is given,
	put "g" spaces between groups of bytes (default is 1 space)

    -L L
	set the "gutter" between the address column and the first hex
	byte (per line) to be "L" spaces wide (default is 2 spaces)

    -N N
	show "N" bytes per line (default is 16 bytes)

    -R R
	set the "gutter" between the last hex byte (per line) and the
	first ASCII byte (again, per line) to be "R" spaces wide
	(default is 2 spaces)

CAVEATS
    The options "-B" and "-g", if used together, can produce unexpected
    results; it may help to remember that both options specify numbers
    of bytes, so if used together, the argument to "-G" should probably
    be larger than and a multiple of the argument given to "-B". Also,
    either option by itself will change the length of the output line,
    possibly making it wider than desired.

EXIT STATUS
    hd exits with 0 on success, 1 on any error (including bad options).

SEE ALSO
    od(1)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====	*/


/**
*** Copyright, 2001 - 2010, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
